# std
import sys
from typing import List, Tuple

# internal
from laz.utils.errors import LazRuntimeError
from laz.cli.parser import parser
from laz.utils.load import load
from laz.utils.logging import get_logger
from laz.utils import log
from laz.model.runner import Runner


def root():
    parser.add_argument('--reverse', '-r', action='store_true', default=False, help='Run targets in reverse order.')

    cli_args, laz_args = _split_args()
    if len(laz_args) == 0:
        raise LazRuntimeError('No path provided')
    elif len(laz_args) == 1:
        log.debug('Running default action')
        laz_args += ['default']

    cli_args = parser.parse_args(cli_args)
    get_logger(verbosity=cli_args.verbose)
    root_node = load()
    runner = Runner(root_node, cli_args, laz_args)
    runner.run()


def _split_args() -> Tuple[List[str], List[str]]:
    for i, s in enumerate(sys.argv):
        if i > 0 and not s.startswith('-'):
            return sys.argv[1:i], sys.argv[i:]
    return sys.argv[1:], []



# std
from typing import List, Union

# internal
from laz.utils import log
from laz.model.configuration import Configuration
from laz.model.target import Target


class Plugin:
    context: Union[Configuration, Target]

    def __init__(self, context: Union[Configuration, Target]):
        self.context = context

    def __init_subclass__(cls, **kwargs):
        PLUGINS.append(cls)
        cls.before_all = log_method(cls.before_all)
        cls.before_target = log_method(cls.before_target)
        cls.after_target = log_method(cls.after_target)
        cls.after_all = log_method(cls.after_all)

    def before_all(self):
        raise NotImplementedError

    def before_target(self):
        raise NotImplementedError

    def after_target(self):
        raise NotImplementedError

    def after_all(self):
        raise NotImplementedError


def log_method(method):
    class_name = method.__qualname__.split('.')[0]
    def wrapper(self, *args, **kwargs):
        if class_name != 'Plugin':
            log.debug(f'Running {self.__class__.__name__} {method.__name__}')
        return method(self, *args, **kwargs)
    return wrapper


PLUGINS: List[type(Plugin)] = []



# std
import argparse
from typing import List

# internal
from laz.utils import log
from laz.utils.contexts import in_dir
from laz.model.tree import Node
from laz.model.path import Path
from laz.model.resolver import Resolver
from laz.model.configuration import Configuration
from laz.model.target import Target
from laz.model.act import Act
from laz.plugins.plugin import PLUGINS


class Runner:

    def __init__(self, root_node: Node, cli_args: argparse.Namespace, args: List[str]):
        self.root_node = root_node
        self.cli_args = cli_args
        self.path = Path(args[0])
        self.args = args[1:]
        self.root_node.configuration.push({
            'path': args[0],
            'args': args[1:],
        })

    def resolve(self) -> List[Target]:
        resolver = Resolver(self.root_node, self.path)
        targets = resolver.resolve()
        if self.cli_args.reverse:
            targets.reverse()
        return targets

    def run(self):
        self.load_plugins(self.root_node.configuration)
        self.before_all(self.root_node.configuration)
        targets = self.resolve()
        for target in targets:
            with in_dir(target.data['dirpath']):
                self.before_target(target)
                args = ' '.join(target.data['args'])
                act = Act.new(target, args=args)
                act.act()
                self.after_target(target)
        self.after_all(self.root_node.configuration)

    @staticmethod
    def load_plugins(configuration: Configuration):
        from importlib import import_module
        default_plugins = ['laz.plugins.jinja']
        configured_plugins = configuration.data.get('plugins', [])
        plugins = default_plugins + configured_plugins
        for import_path in plugins:
            import_module(import_path)

    @staticmethod
    def before_all(configuration: Configuration):
        log.debug(f'Running before_all hooks')
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(configuration)
                plugin.before_all()
            except NotImplementedError:
                pass

    @staticmethod
    def before_target(target: Target):
        log.debug(f'Running before_target hooks')
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(target)
                plugin.before_target()
            except NotImplementedError:
                pass

    @staticmethod
    def after_target(target: Target):
        log.debug(f'Running after_target hooks')
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(target)
                plugin.after_target()
            except NotImplementedError:
                pass

    @staticmethod
    def after_all(configuration: Configuration):
        log.debug(f'Running after_all hooks')
        for Plugin in PLUGINS:
            try:
                plugin = Plugin(configuration)
                plugin.after_all()
            except NotImplementedError:
                pass



# std
from typing import Dict, List, Union

AtomicData = Union[type(None), bool, int, str]
ListData = List[Union[AtomicData, 'ListData', 'DictData']]
DictData = Dict[str, Union[AtomicData, ListData, 'DictData']]
Data = Union[AtomicData, ListData, DictData]



# std
from copy import deepcopy
from typing import Optional as Opt

# external
from jinja2 import Template

# internal
from laz.utils.errors import LazTypeError
from laz.utils.types import AtomicData, Data, DictData, ListData
from laz.plugins.plugin import Plugin


class JinjaPlugin(Plugin):

    def before_target(self):

        self.context.push(_evaluate(self.context.data))


def _evaluate(data: Data, variables: Opt[DictData] = None) -> Data:
    if variables is None and isinstance(data, dict):
        variables = data
    data = deepcopy(data)
    return _expand(data, variables)


def _expand(data: Data, variables: DictData) -> Data:
    if isinstance(data, dict):
        return _expand_dict(data, variables)
    elif isinstance(data, list):
        return _expand_list(data, variables)
    elif isinstance(data, (type(None), bool, int, str)):
        return _expand_atomics(data, variables)
    else:
        raise LazTypeError(f'Cannot expand input of type {type(data)}')


def _expand_dict(data: DictData, variables: DictData) -> DictData:
    for key in data.keys():
        data[key] = _expand(data[key], variables)
    return data


def _expand_list(data: ListData, variables: DictData) -> ListData:
    for i in range(len(data)):
        data[i] = _expand(data[i], variables)
    return data


def _expand_atomics(data: AtomicData, variables: DictData) -> AtomicData:
    if isinstance(data, str):
        template = Template(data)
        rendered = template.render(**variables)
        return rendered
    else:
        return data



# std
from typing import List, Optional as Opt

# internal
from laz.utils.errors import LazValueError
from laz.utils.contexts import in_dir
from laz.model.action import Action
from laz.model.target import Target


class Act:

    def __init__(self, target: Target, action: Action):
        self.target = target
        self.action = action

    def act(self):
        with in_dir(self.target.data['dirpath']):
            self.action.run()

    @classmethod
    def new(cls, target: Target, action: Opt[Action] = None, args: Opt[str] = None):
        if action is None and args is None:
            raise LazValueError('Must provide action or args')
        if action is not None:
            return Act(target, action)
        else:
            action = Action.new(target, args)
            return Act(target, action)



# std
import os

# internal
from laz.model.configuration import Configuration
from laz.model.tree import Node
from laz.model.target import Target


def get_target_configuration(node: Node, target_name: str) -> Target:
    target = Target(target_name)
    node_configuration = get_node_configuration(node)
    target.push({'target': {'name': target_name}})
    target.push(node_configuration.data)
    target.push(node_configuration.get_target(target_name).data)
    return target


def get_node_configuration(node: Node) -> Configuration:
    nodes = node.root_path()
    data = {
        'dirpath': os.path.dirname(node.configuration.filepath),
        'filepath': node.configuration.filepath,
        'name': node.configuration.name,
    }
    conf = Configuration(node.configuration.id, **data)
    for node in nodes:
        conf.push(node.configuration.data)
    return conf



class Path:
    sep = '/'

    def __init__(self, value: str):
        self.value = value

    def __repr__(self):
        return self.value

    @property
    def base(self) -> str:
        return self.sep.join(self.value.split(self.sep)[:-1])

    @property
    def target(self) -> str:
        return self.value.split(self.sep)[-1]

    @property
    def base_pattern(self) -> str:
        if self.base == '':
            return '**'
        else:
            return f'**/{self.base}/**'



# std
import sys
from typing import Optional as Opt

# internal
from laz.utils import log


def main():
    log.debug(sys.argv[1:])
    subcommand = _subcommand()
    if len(sys.argv) > 1 and sys.argv[1] in ('help', '--help', '-h'):
        from laz.cli.subcommands.help import help
        help()
    elif len(sys.argv) == 2 and (sys.argv[1] == 'version' or sys.argv[1] == '--version'):
        from laz.cli.subcommands.version import version
        version()
    elif subcommand == 'init':
        from laz.cli.subcommands.init import init
        init()
    elif subcommand == 'tree':
        from laz.cli.subcommands.tree import tree
        tree()
    else:
        from laz.cli.root import root
        root()


def _subcommand() -> Opt[str]:
    for i, s in enumerate(sys.argv):
        if i > 0 and not s.startswith('-'):
            return sys.argv[i]


if __name__ == '__main__':
    sys.argv = [
        'laz',
        '-vvv',
        'pqr/dev',
        'echo',
        '{{ target.name }}',
    ]
    main()



# std
from __future__ import annotations
import json
import yaml
from yaml.loader import SafeLoader

# internal
from laz.utils.types import Data
from laz.model.serializable import Serializable
from laz.model.stackable import Stackable


class BaseObject(Serializable, Stackable):

    def __init__(self, id: str, **data: Data):
        self.id = id
        self.data = data

    def json(self) -> str:
        return json.dumps(self.data, indent=2, default=str)

    def serialize(self) -> str:
        return yaml.dump(self.data)

    @classmethod
    def deserialize(cls, id: str, serialized: str) -> BaseObject:
        data = yaml.load(serialized, Loader=SafeLoader) or {}
        return cls(id, **data)



# std
from fnmatch import fnmatch
from typing import List

# internal
from laz.model.configuration import Configuration
from laz.model.path import Path
from laz.model.target import Target
from laz.model.tree import Node
from laz.utils.node import get_node_configuration, get_target_configuration


class Resolver:

    def __init__(self, node: Node, path: Path):
        self.node = node
        self.path = path

    def resolve(self) -> List[Target]:
        nodes = self.resolve_nodes()
        return self.resolve_targets(nodes)

    def resolve_nodes(self) -> List[Node]:
        nodes: List[Node] = []
        for node in self.node.nodes():
            name_path = self.get_name_path(node)
            if fnmatch(name_path, self.path.base_pattern):
                nodes.append(node)
        return nodes

    def resolve_targets(self, nodes: List[Node]) -> List[Target]:
        configurations = []
        for node in nodes:
            node_configuration = get_node_configuration(node)
            for target_name in node_configuration.data.get('targets', {}).keys():
                if fnmatch(target_name, self.path.target):
                    target_configuration = get_target_configuration(node, target_name)
                    configurations.append(target_configuration)
        return configurations

    @staticmethod
    def get_name_path(node: Node) -> str:
        nodes = node.root_path()
        names = [node.configuration.name for node in nodes]
        return Path.sep.join(names + [''])



def flatten(to_flatten) -> list:
    flattened = []
    if isinstance(to_flatten, list):
        for item in to_flatten:
            flattened.extend(flatten(item))
    else:
        flattened.append(to_flatten)
    return flattened


def compact(l: list) -> list:
    return [item for item in l if item is not None]



# internal
from laz.internal.meta import version


HELP = f'''Laz version {version}

Usage:
  laz [-h] [-v] <command> [<arg1>] ... [<argN>]

Global Options:
  -h (--help)        Display help message
  -v (--verbose)     Increase verbosity of log messages

Available Commands:
  help               Display help message
  init               Initialize a directory with an example laz.yml
  <path>             Configuration/Target path to run arguments against
  tree               Display tree structure of laz.yml files in current project
  version            Display current Laz package version

Online Documentation: https://joshwycuff.github.io/py-laz/
'''


def help():
    print(HELP)



# std
import os
from typing import List, Optional as Opt, Union

# internal
from laz.utils.errors import LazRuntimeError
from laz.utils import log
from laz.model.configuration import Configuration
from laz.model.tree import Node
from laz.utils.funcs import compact, flatten


def load(dirpath: str = '.') -> Node:
    rootpath = _find_rootpath(dirpath)
    root_node = _build_tree(rootpath)
    return root_node


def _find_rootpath(dirpath: str) -> str:
    rootpath = None
    dirpath = os.path.abspath(dirpath)
    parts = dirpath.split(os.path.sep)
    for i in range(len(parts), 0, -1):
        dirpath = os.path.join(os.path.sep, *parts[:i])
        filepath = os.path.join(dirpath, 'laz.yml')
        if os.path.isfile(filepath):
            rootpath = dirpath
    if rootpath is None:
        raise LazRuntimeError('Could not find root laz.yml')
    log.debug(f'Root laz.yml found: {rootpath}')
    return rootpath


def _build_tree(dirpath: str) -> Opt[Union[Node, List[Node]]]:
    filepath = os.path.join(dirpath, 'laz.yml')
    if os.path.isfile(filepath):
        log.debug(f'laz.yml found: {filepath}')
        node = Node(Configuration.load(filepath))
        children = compact(flatten([_build_tree(d) for d in _listdirs(dirpath)]))
        node.add_children(children)
        return node
    else:
        return compact(flatten([_build_tree(d) for d in _listdirs(dirpath)]))


def _listdirs(dirpath: str) -> List[str]:
    dirs: List[str] = []
    entry: os.DirEntry
    for entry in os.scandir(dirpath):
        if entry.is_dir():
            dirs.append(entry.path)
    return sorted(dirs)



# std
from __future__ import annotations
import os
from typing import List

# internal
from laz.model.base import BaseObject
from laz.model.target import Target


class Configuration(BaseObject):

    @property
    def filepath(self):
        return self.id

    @property
    def name(self):
        return os.path.basename(os.path.dirname(self.filepath))

    @property
    def target_names(self) -> List[str]:
        return list(self.data.get('targets', {}).keys())

    def get_target(self, name: str) -> Target:
        return Target(name, **self.data.get('targets', {}).get(name, {}) or {})

    @classmethod
    def load(cls, filepath: str) -> Configuration:
        with open(filepath, 'r') as fh:
            return Configuration.deserialize(filepath, fh.read())



# std
from typing import List
# internal
from laz.utils.logging import get_logger
from laz.cli.parser import parser
from laz.utils.load import load
from laz.model.tree import Node
from laz.utils.iterator import Iterator
from laz.utils.node import get_node_configuration
from laz.model.resolver import Resolver


subparsers = parser.add_subparsers(required=True)

parser_tree = subparsers.add_parser('tree')
parser_tree.add_argument('--show-targets', '-t', action='store_true', default=False, help='Show targets.')
parser_tree.add_argument('--show-paths', '-p', action='store_true', default=False, help='Show paths.')

# config
SHOW_TARGETS = False
SHOW_PATHS = False


def tree():
    global SHOW_TARGETS, SHOW_PATHS
    args = parser.parse_args()
    get_logger(verbosity=args.verbose)
    SHOW_TARGETS = args.show_targets
    SHOW_PATHS = args.show_paths
    root = load()
    print(root.configuration.name)
    if args.show_targets:
        _targets(root, '', len(root.children) == 0)
    for child in Iterator(root.children):
        _tree(child.value, '', child.is_last)


def _tree(
        node: Node,
        prefix: str = '',
        last: bool = False,
):
    symbol = '└─' if last else '├─'
    path = f'({Resolver.get_name_path(node)})' if SHOW_PATHS else ''
    print(f'{prefix}{symbol} {node.configuration.name}/ {path}')

    if SHOW_TARGETS:
        _targets(node, prefix, last)

    for child in Iterator(node.children):
        _tree(child.value, prefix + '│  ', child.is_last)


def _targets(node: Node, prefix: str, last: bool):
    prefix = prefix + '   ' if last else prefix + '│  '
    for target_name in Iterator(_get_target_names(node)):
        symbol = '└─' if target_name.is_last and len(node.children) == 0 else '├─'
        print(f'{prefix}{symbol} {target_name.value}')


def _get_target_names(node: Node) -> List[str]:
    configuration = get_node_configuration(node)
    return configuration.target_names
